package prompt.roles.data;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.annotations.AuthoredType;
import org.processmining.framework.util.Pair;

import prompt.roles.utils.AuthorsInfo;
import prompt.roles.utils.XLogHelper;

/**
 * This class represents a set of roles associated to a log.
 * 
 * @author Andrea Burattin
 * @author Marco Veluscek
 */
@AuthoredType(
	typeName = "Role Partitioning",
	affiliation = AuthorsInfo.AFFILIATION,
	author = AuthorsInfo.AUTHOR,
	email = AuthorsInfo.EMAIL,
	website = AuthorsInfo.WEBSITE
)
public class RoleSet extends LinkedList<Role> {

	private static final long serialVersionUID = -5264245620120060990L;
	private Map<String, MultiSet<String>> activityOriginatorFrequencies;

	/**
	 * This method adds the provided {@link Dependency} considering both
	 * activities as part of the same role.
	 * 
	 * @param dependency the dependency to add
	 */
	public void sameRoleDependency(Dependency dependency) {
		String sourceActivity = dependency.getSourceActivity();
		String targetActivity = dependency.getTargetActivity();
		
		Role sourceRole = search(sourceActivity);
		Role targetRole = search(targetActivity);
		
		if (sourceRole != null && targetRole != null) {
			if (!sourceRole.equals(targetRole)) {
				add(Role.merge(sourceRole, targetRole));
				remove(sourceRole);
				remove(targetRole);
			}
		}
		if (sourceRole == null && targetRole == null) {
			Role r = new Role();
			r.add(sourceActivity);
			r.add(targetActivity);
			r.addDependency(dependency);
			add(r);
		}
		if (sourceRole == null && targetRole != null) {
			targetRole.add(sourceActivity);
		}
		if (sourceRole != null && targetRole == null) {
			sourceRole.add(targetActivity);
		}
	}
	
	/**
	 * This method adds the provided {@link Dependency} considering the two
	 * activities as belonging to different roles.
	 * 
	 * @param dependency the dependency to add
	 */
	public void handoverDependency(Dependency dependency) {
		String sourceActivity = dependency.getSourceActivity();
		Role sourceRole = search(sourceActivity);
		if (sourceRole == null) {
			sourceRole = new Role();
			sourceRole.add(sourceActivity);
			sourceRole.addDependency(dependency);
			add(sourceRole);
		}
		
		String targetActivity = dependency.getTargetActivity();
		Role targetRole = search(targetActivity);
		if (targetRole == null) {
			targetRole = new Role();
			targetRole.add(targetActivity);
			targetRole.addDependency(dependency);
			add(targetRole);
		}
	}
	
	/**
	 * This method assign the originators of the log to the corresponding
	 * activities and roles.
	 * 
	 * @param log the log to extract the originators from
	 */
	public void assignOriginators(XLog log) {
		for (XTrace t : log) {
			for (XEvent e : t) {
				String activityName = XLogHelper.getName(e);
				String resourceName = XLogHelper.getResource(e);
				Role role = search(activityName);
				if (role != null) {
					role.getOriginators().put(resourceName);
				}
			}
		}
	}
	
	/**
	 * This method returns the role that contains the provided activity
	 * 
	 * @param activity the activity to look for
	 * @return the role containing the provided activities, or <code>null</code>
	 * if such role does not exist
	 */
	public Role search(String activity) {
		for(Role r : this) {
			if (r.contains(activity)) {
				return r;
			}
		}
		return null;
	}
	
	/**
	 * This method continuously aggregates roles as long as the
	 * <code>tau_rho</code> threshold is satisfied.
	 * 
	 * This method corresponds to the Algorithm "Roles Aggregation" (Figure 3)
	 * of the related paper ("Business Models Enhancement through Discovery of
	 * Roles", CIDM 2013).
	 * 
	 * @param tau_rho the <code>tau_rho</code> threshodl
	 */
	public void roleAggregation(double tau_rho) {
		boolean merge = false;
		do {
			Pair<Double, Collection<Pair<Role, Role>>> maximals = bestMergeCandidates();
			if (maximals.getFirst() > tau_rho) {
				Pair<Role, Role> best = selectPartition(maximals.getSecond());
				
				Role union = Role.merge(best.getFirst(), best.getSecond());
				remove(best.getFirst());
				remove(best.getSecond());
				add(union);

				merge = true;
			} else {
				merge = false;
			}
		} while (merge);
	}
	
	/**
	 * This method returns the current set of roles weight
	 * 
	 * @return the weight of the set of roles
	 */
	public double getWeights() {
		double weight = 0.0;
		for (Role role : this) {
			weight += role.getWeight();
		}
		return weight;
	}
	
	/**
	 * This method the set of different and significative <code>rho</code>
	 * values
	 * 
	 * @return the collection of values
	 */
	public Collection<Double> getTrho() {
		Collection<Double> Trho = new HashSet<Double>();
		for (Role r1 : this) {
			for (Role r2 : this) {
				if (!r1.equals(r2)) {
					Trho.add(Role.rho(r1, r2));
				}
			}
		}
		return Trho;
	}
	
	/**
	 * This method finds the best candidate roles to be merged
	 * 
	 * @return the list of candidate roles and the maximal rho
	 */
	private Pair<Double, Collection<Pair<Role, Role>>> bestMergeCandidates() {
		double maxRho = -1;
		Collection<Pair<Role, Role>> maximals = new LinkedList<Pair<Role, Role>>();
		
		for (Role r1 : this) {
			for (Role r2 : this) {
				if (!r1.equals(r2)) {
					double rho = Role.rho(r1, r2);
					if (rho > maxRho) {
						maximals.clear();
					}
					if (rho >= maxRho) {
						maxRho = rho;
						maximals.add(new Pair<Role, Role>(r1, r2));
					}
				}
			}
		}
		
		return new Pair<Double, Collection<Pair<Role, Role>>>(maxRho, maximals);
	}
	
	/**
	 * This method selects the most promising couple of roles for the merging.
	 * 
	 * This method corresponds to the Algorithm "Roles Selection" (Figure 5)
	 * of the related paper ("Business Models Enhancement through Discovery of
	 * Roles", CIDM 2013).
	 * 
	 * @param partitions the collection of partition to select
	 * @return the best couple of roles
	 */
	private Pair<Role, Role> selectPartition(Collection<Pair<Role, Role>> partitions) {
		
		// couple that maximizes the number of merged originators
		Collection<Pair<Role, Role>> maximalsOriginators = new LinkedList<Pair<Role, Role>>();
		int maxOriginators = 0;
		for (Pair<Role, Role> p : partitions) {
			int mergedOriginatos = MultiSet.union(p.getFirst().getOriginators(), p.getSecond().getOriginators()).size();
			if (mergedOriginatos > maxOriginators) {
				maximalsOriginators.clear();
			}
			if (mergedOriginatos >= maxOriginators) {
				maxOriginators = mergedOriginatos;
				maximalsOriginators.add(p);
			}
		}
		
		// couple that maximizes the number of merged activities
		List<Pair<Role, Role>> maximalsActivities = new LinkedList<Pair<Role, Role>>();
		int maxActivities = 0;
		for (Pair<Role, Role> p : maximalsOriginators) {
			int activities = p.getFirst().size() + p.getSecond().size();
			if (activities > maxActivities) {
				maximalsActivities.clear();
			}
			if (activities >= maxActivities) {
				maxActivities = activities;
				maximalsActivities.add(p);
			}
		}
		
		// lexicographical order
		Collections.sort(maximalsActivities, new Comparator<Pair<Role, Role>>() {
			public int compare(Pair<Role, Role> o1, Pair<Role, Role> o2) {
				String s1 = Role.stringRepresentation(o1.getFirst(), o1.getSecond());
				String s2 = Role.stringRepresentation(o2.getFirst(), o2.getSecond());
				return s1.compareTo(s2);
			}
		});
		
		return maximalsActivities.get(0);
	}
	
	/**
	 * 
	 * @param activityOriginatorFrequencies
	 */
	public void setActivityOriginatorFrequencies(Map<String, MultiSet<String>> activityOriginatorFrequencies) {
		this.activityOriginatorFrequencies = activityOriginatorFrequencies;
	}
	
	/**
	 * 
	 * @return
	 */
	public Map<String, MultiSet<String>> getActivityOriginatorFrequencies() {
		return this.activityOriginatorFrequencies;
	}
	
	/**
	 * This method saves a role set into a file
	 * 
	 * @param roleSet the role set to save
	 * @param destinationFile the destination file of the role set
	 */
	public static void saveToFile(RoleSet roleSet, File destinationFile) {
		ObjectOutputStream outputStream = null;
		try{
			outputStream = new ObjectOutputStream(new FileOutputStream(destinationFile.getAbsoluteFile()));
			outputStream.writeObject(roleSet);
		} catch (IOException ex) {
			ex.printStackTrace();
		} finally {
			try {
				if (outputStream != null) {
					outputStream.flush();
					outputStream.close();
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	/**
	 * This method loads a role set from a file
	 * 
	 * @param sourceFile the file containing the role set
	 * @return the role set loaded
	 */
	public static RoleSet loadFromFile(File sourceFile) {
		RoleSet data1 = null;
		ObjectInputStream in = null;
		try {
			in = new ObjectInputStream(new FileInputStream(sourceFile));
			data1 = (RoleSet) in.readObject();
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			try {
				if (in != null) {
					in.close();
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
		return data1;
	}
	
	@Override
	public boolean equals(Object other) {
		if (other instanceof RoleSet) {
			RoleSet otherRoleSet = (RoleSet) other;
			if (size() != otherRoleSet.size()) {
				return false;
			}
			for (Role r : this) {
				if (!otherRoleSet.contains(r)) {
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (Role role : this) {
			sb.append(role.toString());
			sb.append("\n");
		}
		return sb.toString();
	}
}
