package prompt.roles.plugins;

import java.util.HashSet;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventType;
import org.processmining.models.graphbased.directed.bpmn.elements.Flow;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;
import org.processmining.models.graphbased.directed.bpmn.elements.Swimlane;
import org.processmining.models.graphbased.directed.bpmn.elements.SwimlaneType;

import prompt.roles.data.Role;
import prompt.roles.data.RoleSet;
import prompt.roles.utils.AuthorsInfo;

/**
 * This class contains the plugin that is used to decorate a BPMN model with the
 * provided set of roles
 * 
 * @author Andrea Burattin
 */
public class DecorateBPMNModel {

	/**
	 * This method is the actual plugin
	 * 
	 * @param context
	 * @param model
	 * @param roles
	 * @return
	 */
	@Plugin(
		name = "Decorate a BPMN Model with a Role Partition",
		parameterLabels = { "A BPMN model", "A roles partition" },
		returnLabels = { "The decorated BPMN model" },
		returnTypes = { BPMNDiagram.class },
		userAccessible = true
	)

	@UITopiaVariant(
		affiliation = AuthorsInfo.AFFILIATION,
		author = AuthorsInfo.AUTHOR, 
		email = AuthorsInfo.EMAIL,  
		pack = AuthorsInfo.PACK, 
		website = AuthorsInfo.WEBSITE
	)
	public static BPMNDiagram decorateModel(PluginContext context, BPMNDiagram model, RoleSet roles) {
		int i = 1;
		Swimlane s0 = model.addSwimlane("Test 0", null, SwimlaneType.POOL);
		
		// no roles available
		if (roles.size() == 0) {
			Swimlane s = model.addSwimlane("Role 1", s0, SwimlaneType.LANE);
			for (Activity a : model.getActivities()) {
				a.setParentSwimlane(s);
			}
			for(Gateway g : model.getGateways()) {
				g.setParentSwimlane(s);
			}
			for(Event e : model.getEvents()) {
				e.setParentSwimlane(s);
			}
			
			return model;
		}
		
		// there is at least one role
		for (Role r : roles) {
			Swimlane s = model.addSwimlane("Role " + i++, s0, SwimlaneType.LANE);
			for (String awo : r) {
				try {
					Activity a = getActivity(model, awo);
					s.addChild(a);
					a.setParentSwimlane(s);
					
					// also pre-decorate all outgoing gateways/event and incoming event
					for (BPMNNode n : getOutgoing(model, a)) {
						if (n instanceof Gateway || n instanceof Event) {
							n.setParentSwimlane(s);
						}
					}
					for (BPMNNode n : getIncoming(model, a)) {
						if (n instanceof Event) {
							n.setParentSwimlane(s);
						}
					}
				} catch (Exception e) { /* no activity w/ that name, skip */ }
			}
		}
		
		for (Activity a : model.getActivities()) {
			if (a.getParentSwimlane() == null) {
				a.setParentSwimlane(s0);
			}
		}
		
//		Set<Flow> setOfFlows = new HashSet<Flow>();
//		setOfFlows.addAll(model.getFlows());
//		for (Flow f : setOfFlows) {
//			BPMNNode source = f.getSource();
//			BPMNNode target = f.getTarget();
//			model.removeEdge(f);
//			model.addFlow(source, target, s0, f.getLabel());
//		}
		
		for(Gateway g : model.getGateways()) {
			if (g.getParentSwimlane() == null) {
				BPMNNode previous = g;
				Set<BPMNNode> incoming = null;
				Swimlane s = null;
				// first: check all incoming nodes
				do {
					incoming = getIncoming(model, previous);
					if (incoming.size() == 0) {
						break;
					}
					previous = incoming.iterator().next();
					s = previous.getParentSwimlane();
				} while(s == null);
				// if no incoming nodes have been found, try with outgoing
				if (s == null) {
					do {
						incoming = getOutgoing(model, previous);
						if (incoming.size() == 0) {
							break;
						}
						previous = incoming.iterator().next();
						s = previous.getParentSwimlane();
					} while(s == null);
				}
				if (s != null) {
					g.setParentSwimlane(s);
				}
			}
		}
		
		for(Event e : model.getEvents()) {
			if (e.getParentSwimlane() == null) {
				e.setParentSwimlane(s0);
			}
			if (e.getEventType() == EventType.START) {
				e.setParentSwimlane(getOutgoing(model, e).iterator().next().getParentSwimlane());
			}
			if (e.getEventType() == EventType.END) {
				e.setParentSwimlane(getIncoming(model, e).iterator().next().getParentSwimlane());
			}
		}
		
		return model;
	}
	
	/**
	 * This method returns the model activity with the given name
	 * 
	 * @param model the BPMN model
	 * @param name the name of the activity
	 * @return the activity with the provided name
	 * @throws Exception this exception is thrown if there is no activity with
	 * the given name
	 */
	private static Activity getActivity(BPMNDiagram model, String name) throws Exception {
		for (Activity a : model.getActivities()) {
			String label = a.getLabel();
			if (!label.contains("CLASSIFIED_[") && !label.contains("]_CLASSIFIED")) {
				if (label.contains("+")) {
					label = label.substring(0, label.indexOf("+"));
				}
			}
			if (label.equals(name)) {
				return a;
			}
		}
		throw new Exception("No activity with the given name");
	}
	
	/**
	 * This method returns the set of nodes incoming the provided target
	 * 
	 * @param model the model to analyze
	 * @param target the target node of the dependencies
	 * @return the set of nodes entering the target
	 */
	private static Set<BPMNNode> getIncoming(BPMNDiagram model, BPMNNode target) {
		Set<BPMNNode> set = new HashSet<BPMNNode>();
		for(Flow f : model.getFlows()) {
			if (f.getTarget().equals(target)) {
				set.add(f.getSource());
			}
		}
		return set;
	}
	
	/**
	 * This method returns the set of nodes outgoing from the provided source
	 * 
	 * @param model the model to analyze
	 * @param source the source node of the dependencies
	 * @return the set of nodes exiting the target
	 */
	private static Set<BPMNNode> getOutgoing(BPMNDiagram model, BPMNNode source) {
		Set<BPMNNode> set = new HashSet<BPMNNode>();
		for(Flow f : model.getFlows()) {
			if (f.getSource().equals(source)) {
				set.add(f.getTarget());
			}
		}
		return set;
	}
}
