package prompt.roles.plugins;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Flow;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;

import prompt.roles.utils.AuthorsInfo;

/**
 * This class helps in simplifying a BPMN model in order to extract its
 * dependencies.
 * 
 * @author Andrea Burattin
 */
public class PreprocessModel {

	private BPMNDiagram model = null;
	
	@Plugin(
		name = "Preprocess Model",
		parameterLabels = { "Process Model", },
		returnLabels = { "Preprocessed Model" },
		returnTypes = { BPMNDiagram.class },
		userAccessible = true
	)

	@UITopiaVariant(
		affiliation = AuthorsInfo.AFFILIATION,
		author = AuthorsInfo.AUTHOR, 
		email = AuthorsInfo.EMAIL,  
		pack = AuthorsInfo.PACK, 
		website = AuthorsInfo.WEBSITE
	)
	public BPMNDiagram preprocess(PluginContext context, BPMNDiagram diagram) {
		return preprocess(diagram);
	}
	
	/**
	 * The actual method that performs the preprocessing
	 * 
	 * @param diagram the BPMN 
	 * @return
	 */
	public BPMNDiagram preprocess(BPMNDiagram diagram) {
		model = BPMNDiagramFactory.cloneBPMNDiagram(diagram);
		preprocess();
		return model;
	}
	
	/**
	 * This method pre-processes the BPMN model in order to remove unweanted
	 * behaviors.
	 */
	private void preprocess() {
		// preprocessing for gateways
		Collection<Gateway> gateways = model.getGateways();
		Collection<Gateway> gatewaysToRemove = new HashSet<Gateway>();
		for (Gateway g : gateways) {
			disconnectNode(g);
			gatewaysToRemove.add(g);
		}
		Iterator<Gateway> itGateway = gatewaysToRemove.iterator();
		while(itGateway.hasNext()){
			model.removeGateway(itGateway.next());
		}
		
		// preprocessing for events
		Collection<Event> events = model.getEvents();
		Collection<Event> eventsToRemove = new HashSet<Event>();
		for (Event e : events) {
			disconnectNode(e);
			eventsToRemove.add(e);
		}
		Iterator<Event> itEvent = eventsToRemove.iterator();
		while(itEvent.hasNext()){
			model.removeEvent(itEvent.next());
		}
			
		// preprocessing for un-named activities
		Iterator<Activity> itA = model.getActivities().iterator();
		while(itA.hasNext()){
			Activity a = itA.next();
			if (avoidNode(a.getLabel())) {
				disconnectNode(a);
				itA.remove();
			}
		}
	}
	
	/**
	 * This method indicates whether the provided string refers to a node which
	 * has not to be considered
	 * 
	 * @param label the provided label
	 * @return <code>true</code> if the node has not to be considered,
	 * <code>false</code> otherwise
	 */
	public static boolean avoidNode(String label) {
		return "".equals(label.trim()) || label.contains("tau join  ") || label.contains("tau split  ");
	}
	
	/**
	 * This method fixes the connection of a disconnected node
	 * 
	 * @param nodeToDisconnect the node to disconnect
	 */
	private void disconnectNode(BPMNNode nodeToDisconnect) {
		Set<BPMNNode> sources = getIncoming(nodeToDisconnect);
		Set<BPMNNode> targets = getOutgoing(nodeToDisconnect);
		
		for(BPMNNode source : sources) {
			for(BPMNNode target : targets) {
				model.addFlow(source, target, null);
			}
		}
		
		Set<Flow> toRemove = new HashSet<Flow>();
		for(Flow f : model.getFlows()) {
			if (f.getTarget().equals(nodeToDisconnect) || f.getSource().equals(nodeToDisconnect)) {
				toRemove.add(f);
			}
		}
		Iterator<Flow> itFlow = toRemove.iterator();
		while(itFlow.hasNext()){
			Flow f = itFlow.next();
			if (f.getTarget().equals(nodeToDisconnect) || f.getSource().equals(nodeToDisconnect)) {
				model.removeEdge(f);
			}
		}
	}
	
	/**
	 * This method returns the set of nodes entering the target node
	 * 
	 * @param target
	 * @return
	 */
	private Set<BPMNNode> getIncoming(BPMNNode target) {
		Set<BPMNNode> set = new HashSet<BPMNNode>();
		for(Flow f : model.getFlows()) {
			if (f.getTarget().equals(target)) {
				set.add(f.getSource());
			}
		}
		return set;
	}
	
	/**
	 * This method returns the set of nodes exiting from the source node
	 * 
	 * @param source
	 * @return
	 */
	private Set<BPMNNode> getOutgoing(BPMNNode source) {
		Set<BPMNNode> set = new HashSet<BPMNNode>();
		for(Flow f : model.getFlows()) {
			if (f.getSource().equals(source)) {
				set.add(f.getTarget());
			}
		}
		return set;
	}
}
