package prompt.roles.plugins;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.collections15.map.HashedMap;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;

import prompt.roles.data.Dependency;
import prompt.roles.data.DependencySet;
import prompt.roles.data.MultiSet;
import prompt.roles.data.RoleSet;
import prompt.roles.utils.AuthorsInfo;
import prompt.roles.utils.Configuration;
import prompt.roles.utils.DependencyExtractor;
import prompt.roles.utils.Entropy;
import prompt.roles.utils.RolesExtractor;

/**
 * This class contains the main swimlane discovery plugin
 * 
 * @author Andrea Burattin
 */
public class SwimlaneDiscovery {

	/**
	 * This method contains the swimlane discovery plugin
	 * 
	 * @param context
	 * @param diagram
	 * @param log
	 * @return
	 */
	@Plugin(
		name = "Discover Swimlanes",
		parameterLabels = { "Process Model", "Event Log" },
		returnLabels = { "Role Set", "Entropy Measure" },
		returnTypes = { RoleSet.class, Double.class },
		userAccessible = true
	)

	@UITopiaVariant(
		affiliation = AuthorsInfo.AFFILIATION,
		author = AuthorsInfo.AUTHOR, 
		email = AuthorsInfo.EMAIL,  
		pack = AuthorsInfo.PACK, 
		website = AuthorsInfo.WEBSITE
	)
	public static Object[] discover(PluginContext context, BPMNDiagram diagram, XLog log) {
		
		TreeMap<Double, RoleSet> results = new TreeMap<Double, RoleSet>();
		
		Map<String, MultiSet<String>> activityOriginatorFrequencies = new HashedMap<String, MultiSet<String>>();
		
		BPMNDiagram diagramCloned = BPMNDiagramFactory.cloneBPMNDiagram(diagram);
		DependencyExtractor de = new DependencyExtractor(diagramCloned);
		
		// This call is the actual iteration through the entire log
		DependencySet wds = de.getWeightedDependencies(log, activityOriginatorFrequencies);
		
		// Checking for NaN dependencies. This check is due to the way
		// dependencies are discovered, starting from the BPMN model: this
		// approach can easily generates too many "false positive" dependencies
		// (for example in case of concatenated gateways).
		Iterator<Dependency> iterator = wds.iterator();
		while (iterator.hasNext()) {
			Dependency d = iterator.next();
			if (Double.isNaN(d.getDegreeOfNoHandover())) {
				iterator.remove();
			}
		}
		
		// From this line, the code is used just to automatically find the
		// optimal configuration parameters.
		
		Collection<Double> Tw = wds.getTw();
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(Tw.size());
		
		System.out.println("Tw = " + Tw);
		
		for (Double tau_w : Tw) {
			RoleSet candidate_roles = new RolesExtractor(wds, log).connectedComponents(tau_w);
			Collection<Double> Trho = candidate_roles.getTrho();
			for (Double tau_rho : Trho) {
				RolesExtractor re2 = new RolesExtractor(wds, log);
				RoleSet roles = re2.extractRoles(tau_w, tau_rho);
				
				double entropy = Entropy.computeEntropy(roles, log, Configuration.ADD_USER_FREQUENCY_TO_ENTROPY);
				
				if (!results.values().contains(roles)) {
					results.put(entropy, roles);
				}
			}
			context.getProgress().inc();
		}
		
		double bestEntropy = Double.MAX_VALUE;
		RoleSet bestRole = null;
		
		int i = 1;
//		System.out.println("ROLES:");
		for(Double costs : results.keySet()) {
			if (bestRole == null) {
				bestRole = results.get(costs);
				bestEntropy = costs;
			}
			System.out.println("#" + i++ + "\t entropy = " + costs);
			System.out.println(results.get(costs));
			System.out.println("=================");
		}
		
		if (bestRole == null) {
			bestRole = new RoleSet();
		}
		
		bestRole.setActivityOriginatorFrequencies(activityOriginatorFrequencies);
		
		System.out.println("BEST ROLES (h = " + bestEntropy + ")");
		System.out.println("==========");
		System.out.println(bestRole);
		
		return new Object[] {bestRole, bestEntropy};
	}
}
