package prompt.roles.utils;

import org.deckfour.xes.model.XLog;

import prompt.roles.data.Dependency;
import prompt.roles.data.DependencySet;
import prompt.roles.data.RoleSet;

/**
 * This class contains the component responsible for the identification of roles
 * 
 * @author Andrea Burattin
 */
public class RolesExtractor {

	private DependencySet dependencies;
	private XLog log;
	private RoleSet components;
	
	/**
	 * Basic class constructor
	 * 
	 * @param dependencies the set of dependencies
	 * @param log the reference log
	 */
	public RolesExtractor(DependencySet dependencies, XLog log) {
		this.dependencies = dependencies;
		this.log = log;
	}
	
	/**
	 * This method extract the roles, as the connected components, after cutting
	 * the edges with weight below the <code>tau_w</code> threshold.
	 * 
	 * @param tau_w the weight threshold
	 * @return the set of roles as the connected components after cutting the
	 * edges
	 */
	public RoleSet connectedComponents(double tau_w) {
		if (components == null) {
			components = new RoleSet();
			
			// split roles when degree minor to tau_w
			for (Dependency d : dependencies) {
				if (d.getDegreeOfNoHandover() < tau_w) {
					components.handoverDependency(d);
				} else {
					components.sameRoleDependency(d);
				}
			}
			
			// assign originators to each role
			components.assignOriginators(log);
		}
		return components;
	}

	/**
	 * This method perform the complete roles extraction. Specifically, it
	 * performs both the identification of the connected components and the
	 * aggregation according to the <code>tau_rho</code>.
	 *  
	 * @param tau_w the <code>tau_w</code> threshold
	 * @param tau_rho the <code>tau_rho</code> threshold
	 * @return the set of roles
	 */
	public RoleSet extractRoles(double tau_w, double tau_rho) {
		// get connected components
		RoleSet roles = connectedComponents(tau_w);
		
		// merge roles when rho greater than tau_rho
		roles.roleAggregation(tau_rho);
		
		return roles;
	}
}
